#!/usr/bin/env python
#
# Copyright (c) 2016-2020, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import signal
import sys
import json
import os
import subprocess
import re
import hashlib
import socket
import time
import ipaddress

from datetime import datetime
from abc import ABCMeta, abstractmethod

MAX_SCRIPT_RETRIES = 4
HASH_SALT = 'aY!wB&42YBpSj5p4'

def sigterm_handler(_signo, _stack_frame):
    # Raises SystemExit(0):
    print("Killed, oh no!")
    sys.exit(0)


class CEPCollecting(object):
    __metaclass__ = ABCMeta

    def is_hci(self):
        if os.path.exists("/etc/hci-release"):
            return True
        return False

    def execute_shell(self, command):
        p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        return out

    def service_active(self, service):
        """Return True if service is running/waiting"""
        cmd = '/bin/systemctl is-active %s' % service
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.communicate()
        if proc.returncode == 0:
           return True
        return False

    def service_restart_reload(self, action, service):
        """Return True if service is running/waiting"""
        cmd = '/bin/systemctl %s %s' % (action, service)
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.communicate()
        if proc.returncode == 0:
           return True
        return False

    def is_exist(self, service):
        """Return True if service is exist"""
        cmd = '/bin/systemctl list-unit-files'
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout_list = proc.communicate()[0]
        pat_match = re.search(service, stdout_list)
        if pat_match is not None:
           return True
        return False

    def container_exist(self, container):
        """Return True if container is exist"""
        cmd = '/usr/bin/docker ps -q -f name=%s' % (container)
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.communicate()
        if proc.returncode == 0:
           return True
        return False


    def port_knock(self, ip, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # 1 sec timeout
        s.settimeout(1)
        try:
                s.connect((ip, int(port)))
                s.shutdown(socket.SHUT_RDWR)
                return True
        except:
                return False
        finally:
                s.close()


    def prepare_tags(self, d):
        if isinstance(d, dict):
            for key in d.keys():
                self.prepare_tags(d[key])
                new_tag = key.replace(" ", "")
                if key[0].isdigit():
                    new_tag = "_" + new_tag
                d[new_tag] = d.pop(key)
        else:
            return

    def prepare(self):
        self.label = self.__class__.__name__
        if len(sys.argv) != 3:
            raise ValueError("Bad script argument")
            sys.exit(1)
        signal.signal(signal.SIGTERM, sigterm_handler)
        self.label = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        self.result = {}

    def serialize(self):
        t = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.result["CollectDate"] = t
        with open(os.path.join(sys.argv[1], self.label + ".json"),
                  "w") as f:
            json.dump(self.result, f)

    def anonymize(self):
        if self.is_hci():
            # NOTE: check for circular references
            try:
                json.dumps(self.result, check_circular=True)
            except:
                return
            self.anonymize_recursive(self.result)

    def anonymize_recursive(self, value, level = 0):
        if isinstance(value, unicode) or isinstance(value, str):
            value = self.hide_hwaddrs(value)
            value = self.hide_ips(value)
        elif isinstance(value, dict):
            skip = []
            if level == 0:
                skip = self.skip_anonymization()
            for i in value.keys():
                if skip and i in skip:
                    continue
                value[i] = self.anonymize_recursive(value[i], level + 1)
        elif isinstance(value, list):
            for i in range(len(value)):
                value[i] = self.anonymize_recursive(value[i], level + 1)
        elif isinstance(value, tuple):
            tuple_list = []
            for i in range(len(value)):
                tuple_list.append(self.anonymize_recursive(value[i], level + 1))
            value = tuple(tuple_list)
        return value

    def skip_anonymization(self):
        return ['packages']

    @abstractmethod
    def fill_results(self):
        pass

    def run(self):
        try:
            self.prepare()
            self.fill_results()
            self.anonymize()
            self.serialize()
        except Exception as e:
            print(e)
            d = "/vz/tmp/cep/"
            if len(sys.argv) > 1 and os.path.isdir(sys.argv[1]):
                d = sys.argv[1]

            p = os.path.join(d, self.label + "-fail.log")
            try:
                with open(p, 'a'):
                    os.utime(p, None)
            except:
                pass

    def hide_hwaddrs(self, val):
        macs = re.findall(r'([^\.\:\w\-]|\A)([a-f\d]{2}:[a-f\d]{2}:[a-f\d]{2}:[a-f\d]{2}:[a-f\d]{2}:[a-f\d]{2})([^\.\:\w\-]|\Z)', val, flags=re.IGNORECASE)
        for mac in macs:
            hashed = hashlib.sha512(mac[1] + HASH_SALT).hexdigest()[0:32].upper()
            val = val.replace(mac[1], hashed)
        return val

    def hide_ips(self, val):
        words = re.split('([^\.\:\w\-\/])', val)

        for i in range(len(words)):
            port = None
            if ':' in words[i]:
                m = re.search(r'\A([\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}):(\d+)\Z', words[i])
                if m:
                    words[i] = m.group(1)
                    port = m.group(2)

            suffix = None
            m = re.search(r'\A(.+)([\-\_\.]+)\Z', words[i])
            if m:
                words[i] = m.group(1)
                suffix = m.group(2)

            try:
                res = ipaddress.ip_address(unicode(words[i]))
                if type(res).__name__ == 'IPv6Address':
                    words[i] = 'XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:XXXX'
                else:
                    words[i] = 'XXX.XXX.XXX.XXX'
                port = 'XXX'
            except:
                pass

            if port is not None:
                words[i] += ':' + port
            if suffix is not None:
                words[i] += suffix

        return ''.join(words)

    def dir_size(self, path):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(path):
            for f in filenames:
                total_size += os.path.getsize(dirpath + '/' + f)

        return total_size

    def read_ini_file(self, path):
        if not os.path.isfile(path):
            return {}
        regex = re.compile("^([^=\#]+?)\s*=\s*([^\#]*?)$")
        d = dict()
        with open(path, "r") as f:
            for l in f.readlines():
                try:
                    (name, value) = re.findall(regex, l.strip())[0]
                    d[name] = value.strip("\"")
                except:
                    continue
        return d
